#  BSD 3-Clause License
#
#  Copyright (c) 2023, Trixter GmbH
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import inspect
import logging
import re

from collections import OrderedDict
from copy import (
    copy,
    deepcopy
)
import wrapt

import ftrack_api


from .criteria import (
    Criterion,
    Criteria
)
from ..entities.base import (
    Entity,
    EntityCollection,
    EmptyCollection
)
from ..entities.declarations import (
    ForwardDeclareCompare,
    RelationshipDeclaration
)
from ..session import SESSION

from ..configuration import (
    RELATIONSHIPS_RESOLVER,
    LOGGING_NAMESPACE
)
from ..constants import __VERSION__


LOG = logging.getLogger("{}.query".format(LOGGING_NAMESPACE))


class QuerySchema(object):

    def __init__(self, relationships, default_schema="default"):

        primary_keys_to_exclude = ["overrides"]
        primary_entities = relationships.get("entities", {})
        setattr(self, default_schema, {})
        getattr(self, default_schema)["name"] = default_schema
        for key, value in relationships.items():
            if key not in primary_keys_to_exclude:
                getattr(self, default_schema).update({key: value})

        for override in relationships.get("overrides", []):
            overrides = {"entities": {}, "name": override["name"]}
            entities = deepcopy(primary_entities) if override.get("inherit", True) else {}
            if entities:
                overrides["entities"] = entities

            for key, value in override["entities"].items():
                per_entity_relationships = overrides["entities"].get(key, {})
                per_entity_relationships.update(value)
                overrides["entities"][key] = per_entity_relationships

            setattr(self, override["name"], overrides)

    def __getitem__(self, item):
        try:
            return getattr(self, item)
        except AttributeError:
            raise KeyError("Schema '{}' doesn't exist".format(item))


SCHEMA = QuerySchema(RELATIONSHIPS_RESOLVER(api_version=__VERSION__))
_NEGATION_CRITERION_PATTERN = r"^not_by_"


class Query(object):
    """ wraps the ftrack query language

    """

    _known_order = ["ascending", "descending"]

    def __init__(self, entity, session=SESSION, schema=SCHEMA.default):
        self.entity_type = entity()

        self.entity_type.relationship(session=session, schema=schema)

        self.primary_key = self.entity_type.projections[0] if len(self.entity_type.projections) else "id"
        self.projections = []
        self.criteria = []
        self.session = session
        self.schema = schema
        self._valid = False

    def __getattribute__(self, item):
        """ forwarding of criteria methods to the specific entity implementation """

        is_negation = bool(re.match(_NEGATION_CRITERION_PATTERN, item))
        has_criterion = hasattr(Criteria, item)
        if is_negation:
            criterion_counterpart = re.sub(_NEGATION_CRITERION_PATTERN, "by_", item)
            if hasattr(Criteria, criterion_counterpart):
                item = criterion_counterpart
                has_criterion = True

        # check if item is a known criterion
        if has_criterion and item not in dir(type):
            # check if criterion was implemented on entity
            if not hasattr(self.entity_type, item):
                raise Criteria.CriterionNotImplementedError(
                    "Criterion `{}` is not implemented on entity {}.".format(item, self.entity_type)
                )

            return self._magic(getattr(self.entity_type, item), negate=is_negation)

        return super(Query, self).__getattribute__(item)

    def __copy__(self):
        query = Query(self.entity_type.__class__, session=self.session, schema=self.schema)
        query.projections = copy(self.projections)
        query.criteria = copy(self.criteria)
        query.valid = self.valid
        return query

    def __str__(self):
        """ returns the generated query string """

        projections = ", ".join(self.projections or self.entity_type.projections)

        _query = "select {} from {}".format(
                projections,
                self.entity_type.__class__.__name__
            )

        if self.criteria:
            # We sort the criteria based on the distance to the queried entity:
            # distance(parent) = 0
            # distance(parent.parent.name) = 2
            # This will make the query much more efficient.
            _query += " where {}".format(
                    " and ".join(sorted(
                        [_.resolve() for _ in self.criteria],
                        key=lambda x: len(x.split(".")))
                    )
            )

        return _query

    def _get_shared_criterion(self, criterion):
        """ detect if criterion already exists

        Args:
            criterion (Criterion): Criterion object

        Returns:
            bool: True if it can be considered as shared
        """
        _existing = [_ for _ in self.criteria if _.name == criterion.name and _.negate == criterion.negate]
        if _existing:
            if _existing[0] and _existing[0].target == criterion.target:
                return _existing[0]

    def _store_criterion(self, criterion):
        """ detect and merge with same existing criterion

        Args:
            criterion (Criterion): Criterion object

        Returns:

        """
        existing = self._get_shared_criterion(criterion)
        if not existing:
            self.criteria.append(criterion)
        elif existing.args != criterion.args:
            _data = existing.as_dict()
            # append new arguments
            _data["args"] += criterion.args

            index = self.criteria.index(existing)
            self.criteria.remove(existing)
            self.criteria.insert(index, Criterion(**_data))

    def _extract_target(self, args):
        if not args:
            return None, args
        target_cls = args[0]
        if not inspect.isclass(target_cls):
            target_cls = None
            arguments = args
        else:
            arguments = args[1:]
        return target_cls, arguments

    @staticmethod
    def _validate_criteria_signature(func, target, args, kwargs):
        caller_name = func.__name__
        if target and (bool(args) is bool(kwargs)):
            raise Criteria.InvalidArgumentsError(
                "You provided a target, but either NO arguments and NO keyword arguments "
                "or arguments AND keyword arguments in `.{}()`. Both possibilities are "
                "not supported.".format(caller_name)
            )
        elif not target and not args and not kwargs:
            raise Criteria.InvalidArgumentsError(
                "You gave us nothing. NOTHING!!!! `.{}()` is very sad.".format(caller_name)
            )

    def _construct_entity_collection(self, result):
        _cls = self.entity_type.__class__
        if result:
            if isinstance(result, list):
                entities = OrderedDict((_[self.primary_key], Entity(_cls=_cls, ftrack_entity=_)) for _ in result)
            else:
                entities = OrderedDict([(result[self.primary_key], Entity(_cls=_cls, ftrack_entity=result))])

            entitycollection = EntityCollection(_cls=_cls, entities=entities)
            entitycollection.query = Query(_cls, session=self.session, schema=self.schema)

            # we use slicing to create copies of our lists
            # that's how we roll bitch (Dennis said)
            entitycollection.query.criteria = self.criteria[:]
            entitycollection.query.projections = self.projections[:]
            entitycollection.query.valid = True

            return entitycollection
        else:
            return EmptyCollection(_type=_cls())

    def _magic(self, func, negate=False):
        """ wrapper for our forwarded criteria functions

        This allows to append the result of the forwarded function to the
        criteria list, but also returns the Query object itself, to allow
        autocompletion for arbitrary amounts of criteria calls
        (all by_* methods defined in the Criteria class).
        """
        def inner(*args, **kwargs):
            target = None
            arguments = ()

            if args:
                target, arguments = self._extract_target(args)
                self._validate_criteria_signature(func, target, arguments, kwargs)

            criterion = Criterion(
                name=func.__name__,
                args=arguments,
                kwargs=kwargs,
                target=target,
                filter=func,
                negate=negate
            )

            self._store_criterion(criterion)

            return self

        return inner

    @wrapt.decorator
    def _fill_target(wrapped, instance, args, kwargs):
        """ decorator for those terminators which commonly support a target (Entity class)
        """
        target, args = instance._extract_target(args)
        return wrapped(target or instance.entity_type.__class__, *args, **kwargs)
    
    def _set_projections(self, projections):
        if projections:
            # resolve unknown projections
            _projections = []
            for projection in projections:
                # prepare for resolution of something like projections=[Asset]
                if isinstance(projection, ForwardDeclareCompare):
                    projection = RelationshipDeclaration(parent=projection, child="")
                # resolved already wrapped - like projections=[AssetBuild.name]
                if isinstance(projection, RelationshipDeclaration):
                    projection = projection.resolve_path_for(self.entity_type, self.session, self.schema)
                    _projections.extend(projection)
                else:
                    _projections.append(projection)
            # extend default projections
            self.projections = list(set(_projections + self.entity_type.__class__.projections))

    @property
    def valid(self):
        return self._valid

    @valid.setter
    def valid(self, value):
        assert isinstance(value, bool), "Invalid value {} for attribute valid.".format(value)
        self._valid = value

    def get(self, projections=None, limit=None, offset=0, order="ascending", order_by=None):
        """ get the result with limit, offset and specific ordering

        Notes:
            Compared to the default FTrack behavior our order will reflect the
            actual sorting in the end.

        Args:
            limit (int): limited amount of resulting entities
            offset (int): offset
            order (str): "descending" or "ascending"
            order_by (str): field to use for ordering

        Returns:
            Entity subclass: the given entity instance

        """
        assert order in self._known_order, "Unknown specified order '{}'. Supported: {}".format(
            order, ", ".join(self._known_order)
        )
        if projections:
            assert isinstance(projections, (list, tuple, set)), (
                "'projections' expect any of the following types: list, tuple, set. Got: {}".format(type(projections))
            )

        self._set_projections(projections)

        query = str(self)

        if limit:
            query += " limit {}".format(limit)
        if offset:
            query += " offset {}".format(offset)
        if order_by:
            query += " order by {} {}".format(order_by, order)
        if limit and not order_by:
            # TODO: if we query a <dynamic ftrack Recipient object 139776416590416>
            #  which does not have an id this will fail. In this case we need to
            #  sort by a different key
            #  We probably have to dynamically get the primary key and use that
            #  for sorting
            # TEST: Is this solved via the primary key?
            query += " order by {}".format(self.primary_key)

        # In case the same query is reused multiple times, we can lookup a
        # previous result and skip the actual query.
        if self.session.reuse_query_results and query in self.session._terminated_queries:
            LOG.debug("Reuse cached query '{}'".format(query))
            return self.session._terminated_queries[query]

        LOG.info("Performing query: \"{}\"".format(query))
        result = self.session.query(query).all()

        if order_by:
            result = sorted(result, key=lambda x: x[order_by], reverse=(order == "descending"))

        # TODO: if result is empty we need to add the entity type information
        #  to the resulting EmptyCollection
        # TEST: Check if this is already solved.
        entities = self._construct_entity_collection(result)
        entities._session = self.session
        if not entities:
            LOG.warning("Query '{}' returned no results.".format(self))

        # TODO: should we only store the result in case it produced a collection?
        # original query
        self.session._terminated_queries[query] = entities
        # query with original projections
        self.session._terminated_queries[str(entities.as_query(use_ids=True, with_projections=True))] = entities
        # query with exclusive default projections
        self.session._terminated_queries[str(entities.as_query(use_ids=True))] = entities

        return entities

    def get_one(self, projections=None):
        result = self.get(projections=projections, limit=2)
        if len(result) > 1:
            raise ftrack_api.exception.MultipleResultsFoundError("Multiple results found for '{}'".format(self))
        if not result:
            raise ftrack_api.exception.NoResultFoundError("No result found for '{}'".format(self))

        return result

    def get_first(self, projections=None, offset=0):
        return self.get(projections=projections, limit=1, offset=offset)

    def get_all(self, projections=None, limit=None, offset=0, order="ascending", order_by=None):
        return self.get(
            projections=projections,
            limit=limit,
            offset=offset,
            order=order,
            order_by=order_by,
        )
    
    @_fill_target
    def get_inputs(self, target, projections=None):
        in_ids = self.get_all(projections=['incoming_links.from_id']).incoming_links.from_id
        if in_ids:
            inputs = Query(target, session=self.session, schema=self.schema) \
                .by_id(*in_ids) \
                .get_all(projections=projections)
            return inputs
        return EmptyCollection(_type=target(), session=self.session)
    
    @_fill_target
    def get_outputs(self, target, projections=None):
        out_ids = self.get_all(projections=['outgoing_links.to_id']).outgoing_links.to_id
        if out_ids:
            outputs = Query(target, session=self.session, schema=self.schema) \
                .by_id(*out_ids) \
                .get_all(projections=projections)
            return outputs
        return EmptyCollection(_type=target(), session=self.session)



#  BSD 3-Clause License
#
#  Copyright (c) 2023, Trixter GmbH
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from .base import *
from .declarations import *
from ..query import Criteria
from ..query.utils import build_partial_query
from .schematypes import (
    ASSET_TYPES,
    OBJECT_TYPES,
    TASK_TYPES,
    PROJECT_SCHEMAS
)


class AssetVersion(Entity):
    projections = ["id", "asset.name", "version"]

    relationship = Relationship()
    relationship.parent = "asset"

    def pre_create(self, **kwargs):
        """ pre process before create() call

        Validates and modifies passed keyword arguments from create()

        Args:
            **kwargs (): incoming keyword arguments from create(**kwargs)

        Returns:

        """
        super(AssetVersion, self).pre_create(**kwargs)

        assert "task" in kwargs, "task is a required keyword argument"
        assert len(kwargs["task"]) == 1, "AssetVersions have to be linked to a single task. You provided {}.".format(
            len(kwargs["task"])
        )
        assert isinstance(kwargs["task"], EntityCollection), "Argument 'task' expects a value of EntityCollection[Task]"

        kwargs["task"] = list(kwargs["task"].values())[0].ftrack_entity

        return kwargs

    # only for autocompletion
    def create(self, task, **kwargs): pass
    def link_inputs(self, entity_collection): pass
    def link_outputs(self, entity_collection): pass
    def unlink_inputs(self, entity_collection): pass
    def unlink_outputs(self, entity_collection): pass

    @Criteria.supported_targets(Entity)
    def by_name(self, target, *names):
        target = target or self.relationship[Asset]
        return build_partial_query(target, names, "name")

    @Criteria.supported_targets()
    def by_publisher(self, target, *publishers):
        target = target or TargetRelation()
        return build_partial_query(target, publishers, "user.username")

    @Criteria.supported_targets()
    def by_resource_identifier(self, target, *resource_identifiers):
        target = target or self.relationship[ComponentLocation]
        return build_partial_query(target, resource_identifiers, "resource_identifier")

    @Criteria.supported_targets()
    def by_version(self, target, *versions):
        target = target or TargetRelation()
        return build_partial_query(target, versions, "version")

    @Criteria.supported_targets(Task)
    def by_status(self, target, *statuses):
        target = target or TargetRelation()
        return build_partial_query(target, statuses, "status.name")

    @Criteria.supported_targets(Task)
    def by_state(self, target, *states):
        target = target or TargetRelation()
        return build_partial_query(target, states, "status.state.name")

    @Criteria.supported_targets(Task, Asset)
    def by_type(self, target, *types):
        target = target or self.relationship[Asset]
        return build_partial_query(target, types, "type.name")

    @Criteria.supported_targets()
    def by_publish_time(self, target, start=None, end=None):
        query = "("
        if start:
            query += "date >= \"{}\"".format(start)
        if start and end:
            query += " and "
        if end:
            query += "date <= \"{}\"".format(end)
        query += ")"
        return query

    @Criteria.supported_targets()
    def by_outgoing_link(self, target, *ids):
        target = target or TargetRelation("used_in_versions", collection=True)
        return build_partial_query(target, ids, "id")

    @Criteria.supported_targets()
    def by_incoming_link(self, target, *ids):
        target = target or TargetRelation("uses_versions", collection=True)
        return build_partial_query(target, ids, "id")

    @Criteria.supported_targets()
    def by_publish_state(self, target, publish_state):
        target = target or TargetRelation()
        return build_partial_query(target, [publish_state], "is_published")


class TypedContext(Entity):
    projections = ["id", "name"]
    relationship = Relationship()

    # only for autocompletion
    def link_inputs(self, entity_collection): pass
    def link_outputs(self, entity_collection): pass
    def unlink_inputs(self, entity_collection): pass
    def unlink_outputs(self, entity_collection): pass

    @Criteria.supported_targets(Entity)
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")

    @Criteria.supported_targets(AssetVersion, TypedContext)
    def by_status(self, target, *statuses):
        target = target or TargetRelation()
        return build_partial_query(target, statuses, "status.name")

    @Criteria.supported_targets(AssetVersion, TypedContext)
    def by_state(self, target, *states):
        target = target or TargetRelation()
        return build_partial_query(target, states, "status.state.name")

    @Criteria.supported_targets(Asset, TypedContext)
    def by_type(self, target, *types):
        target = target or TargetRelation()
        return build_partial_query(target, types, "type.name")

    @Criteria.supported_targets()
    def by_status_change_time(self, target, start=None, end=None):
        query = "status_changes any ("
        if target:
            query = "{}.{}".format(target.relation, query)
        if start:
            query += "date >= \"{}\"".format(start)
        if start and end:
            query += " and "
        if end:
            query += "date <= \"{}\"".format(end)
        query += ")"
        return query

    @Criteria.supported_targets()
    def by_assignee(self, target, *assignees):
        target = target or TargetRelation()
        return build_partial_query(target, assignees, "assignments.resource.username")

    @Criteria.supported_targets()
    def by_outgoing_link(self, target, *ids):
        target = target or TargetRelation("outgoing_links", collection=True)
        return build_partial_query(target, ids, "to_id")

    @Criteria.supported_targets()
    def by_incoming_link(self, target, *ids):
        target = target or TargetRelation("incoming_links", collection=True)
        return build_partial_query(target, ids, "from_id")

    @Criteria.supported_targets(Project)
    def by_lifespan(self, target, start=None, end=None):
        query = ""
        if target:
            implicit_target = "{}.".format(target.relation)
        else:
            implicit_target = ""

        if start:
            query += "{}start_date >= \"{}\"".format(implicit_target, start)
        if start and end:
            query += " and "
        if end:
            query += "{}end_date <= \"{}\"".format(implicit_target, end)

        return query


class Task(TypedContext):
    relationship = Relationship()

    def pre_create(self, **kwargs):
        """ pre process before create() call

        Validates and modifies passed keyword arguments from create()

        Args:
            **kwargs (): incoming keyword arguments from create(**kwargs)

        Returns:

        """
        super(Task, self).pre_create(**kwargs)

        session = kwargs["session"]
        assert "name" in kwargs, "name is a required keyword argument"
        assert "type" in kwargs, "type is a required argument"

        assert kwargs["type"] in TASK_TYPES.get("types", session), \
            "Unknown task_type '{}'. Known types: {}".format(kwargs["type"], ", ".join(TASK_TYPES.get("types", session).keys()))
        kwargs["type"] = TASK_TYPES.get(kwargs["type"], session)

        return kwargs

    # only for autocompletion
    def create(self, name, task_type): pass


class Shot(TypedContext):
    relationship = Relationship()

    def pre_create(self, **kwargs):
        """ pre process before create() call

        Validates and modifies passed keyword arguments from create()

        Args:
            **kwargs (): incoming keyword arguments from create(**kwargs)

        Returns:

        """
        super(Shot, self).pre_create(**kwargs)

        assert "name" in kwargs, "name is a required keyword argument"
        return kwargs

    # only for autocompletion
    def create(self, name, **kwargs): pass


class Sequence(TypedContext):
    relationship = Relationship()

    def pre_create(self, **kwargs):
        """ pre process before create() call

        Validates and modifies passed keyword arguments from create()

        Args:
            **kwargs (): incoming keyword arguments from create(**kwargs)

        Returns:

        """
        super(Sequence, self).pre_create(**kwargs)

        assert "name" in kwargs, "name is a required keyword argument"
        return kwargs

    def create(self, name, **kwargs): pass


class AssetBuild(TypedContext):
    relationship = Relationship()


class AssetGroup(TypedContext):
    relationship = Relationship()


class Component(Entity):
    projections = ["id", "name"]

    relationship = Relationship()
    relationship.parent = "version"
    
    @Criteria.supported_targets(Entity)
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")

    @Criteria.supported_targets()
    def by_resource_identifier(self, target, *resource_identifiers):
        target = target or self.relationship[ComponentLocation]
        return build_partial_query(target, resource_identifiers, "resource_identifier")

    @Criteria.supported_targets()
    def by_location(self, target, *component_locations):
        target = target or self.relationship[ComponentLocation]
        return build_partial_query(target, component_locations, "location.name")

    @Criteria.supported_targets()
    def by_file_type(self, target, *file_types):
        target = target or TargetRelation()
        return build_partial_query(target, file_types, "file_type")

    @Criteria.supported_targets(AssetVersion)
    def by_version(self, target, *versions):
        target = target or self.relationship[AssetVersion]
        return build_partial_query(target, versions, "version")

    @Criteria.supported_targets()
    def by_system_type(self, target, *system_types):
        target = target or TargetRelation()
        return build_partial_query(target, system_types, "system_type")

    @Criteria.supported_targets()
    def by_size(self, target, minimum=0, maximum=0):
        """Builds a query expression for filtering by size in bytes.

        Args:
            target (NoneType): unused, but required in signature due to decorator.
            minimum (int): minimum size.
            maximum (int): maximum size.

        Returns:
            str: by size partial query.

        """
        query = ""

        if minimum:
            query += "size >= \"{}\"".format(minimum)
        if minimum and maximum:
            query += " and "
        if maximum:
            query += "size <= \"{}\"".format(maximum)

        return query


class FileComponent(Component):
    projections = ["id", "name"]


class ContainerComponent(Component):
    projections = ["id", "name"]


class NoteComponent(Component):
    projections = ["note_id", "component_id", "component"]


class Project(Entity):
    projections = ["id", "name"]

    @Criteria.supported_targets(Entity)
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")

    @Criteria.supported_targets()
    def by_status(self, target, status):
        return "status like {}".format(status)

    # TODO - refactor, other entities use this too
    @Criteria.supported_targets()
    def by_lifespan(self, target, start=None, end=None):
        query = ""
        if target:
            implicit_target = "{}.".format(target.relation)
        else:
            implicit_target = ""

        if start:
            query += "{}start_date >= \"{}\"".format(implicit_target, start)
        if start and end:
            query += " and "
        if end:
            query += "{}end_date <= \"{}\"".format(implicit_target, end)

        return query

    def pre_create(self, **kwargs):
        super(Project, self).pre_create(**kwargs)

        session = kwargs.get("session")
        assert "project_schema" in kwargs, "project_schema is a required keyword argument"
        assert "name" in kwargs, "name is a required keyword argument"
        assert kwargs["project_schema"] in PROJECT_SCHEMAS.get("types", session), \
            "Unknown project_schema '{}'. Known schemas: {}".format(
                kwargs["project_schema"], ", ".join(PROJECT_SCHEMAS.get("types", session).keys())
            )

        kwargs["project_schema"] = PROJECT_SCHEMAS.get(kwargs["project_schema"], session)
        kwargs["no_parent"] = True
        if "full_name" not in kwargs:
            kwargs["full_name"] = kwargs["name"]
        return kwargs

    def create(self, name, project_schema, **kwargs): pass


class ComponentLocation(Entity):
    projections = ["id", "resource_identifier"]

    relationship = Relationship()
    relationship.parent = "component"

    def pre_create(self, **kwargs):
        super(ComponentLocation, self).pre_create(**kwargs)

        if "location" in kwargs:
            kwargs["location"] = list(kwargs["location"].values())[0].ftrack_entity
        else:
            kwargs.setdefault("location_id", "cb268ecc-8809-11e3-a7e2-20c9d081909b")  # unmanaged location

        assert "resource_identifier" in kwargs, "resource_identifier is a required keyword argument"
        return kwargs

    @Criteria.supported_targets(Entity)
    def by_name(self, target, *names):
        target = target or TargetRelation("location")
        return build_partial_query(target, names, "name")

    @Criteria.supported_targets(AssetVersion)
    def by_version(self, target, *versions):
        target = target or self.relationship[AssetVersion]
        return build_partial_query(target, versions, "version")

    @Criteria.supported_targets()
    def by_resource_identifier(self, target, *resource_identifiers):
        target = target or TargetRelation()
        return build_partial_query(target, resource_identifiers, "resource_identifier")


class Asset(Entity):
    projections = ["id", "name"]

    relationship = Relationship()

    @Criteria.supported_targets(Entity)
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")

    @Criteria.supported_targets(TypedContext)
    def by_type(self, target, *types):
        target = target or TargetRelation()
        return build_partial_query(target, types, "type.name")

    def pre_create(self, **kwargs):
        """ pre process before create() call

        Validates and modifies passed keyword arguments from create()

        Args:
            **kwargs (): incoming keyword arguments from create(**kwargs)

        Returns:

        """
        super(Asset, self).pre_create(**kwargs)

        session = kwargs["session"]
        assert "type" in kwargs, "asset_type is a required keyword argument"
        assert "name" in kwargs, "name is a required keyword argument"
        assert kwargs["type"] in ASSET_TYPES.get("types", session), \
            "Unknown asset_type '{}'. Known types: {}".format(kwargs["type"], ", ".join(ASSET_TYPES.get("types", session).keys()))
        kwargs["type"] = ASSET_TYPES.get(kwargs["type"], session)
        return kwargs

    # only for autocompletion
    def create(self, name, type, **kwargs): pass


class Status(Entity):
    projections = ["id", "name"]

    relationship = Relationship()

    # TODO: This needs proper target support; possible targets -> State, Status
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")


class State(Entity):
    projections = ["id", "name"]

    relationship = Relationship()


class AssetVersionLink(Entity):
    projections = ["id", "from_id", "to_id"]


class ProjectSchema(Entity):
    projections = ["id", "name"]


class User(Entity):
    projections = ["id", "username", "is_active"]

    relationship = Relationship()

    @Criteria.supported_targets()
    def by_active_state(self, target, *states):
        target = target or TargetRelation()
        return build_partial_query(target, states, "is_active")

    @Criteria.supported_targets(Entity)
    def by_name(self, target, *names):
        if not target:
            target = TargetRelation()
            return build_partial_query(target, names, "username")
        else:
            return build_partial_query(target, names, "name")


class Timelog(Entity):
    projections =["id", "comment", "start", "duration"]

    relationship = Relationship()
    relationship.parent = "user"

    @Criteria.supported_targets(User, Project)
    def by_name(self, target, *names):
        target = target or TargetRelation()
        if target is self.relationship[User]:
            return build_partial_query(target, names, "username")
        else:
            return build_partial_query(target, names, "name")

    @Criteria.supported_targets()
    def by_lifespan(self, target, start=None, end=None):
        if start:
            query = "start >= \"{}\"".format(start)
        if start and end:
            duration = int((end - start).total_seconds())
            query += " and duration <= \"{}\"".format(duration)
        if end and not start:
            pass

        return query


class UserSecurityRole(Entity):
    projections = ["id", "security_role", "projects"]


class UserSecurityRoleProject(Entity):
    projections = ["id", "project"]


class SecurityRole(Entity):
    projections = ["id", "name", "type"]


class Appointment(Entity):
    projections = ["id", "resource", "resource.name"]

    relationship = Relationship()
    relationship.parent = "context"

    def pre_create(self, **kwargs):
        super(Appointment, self).pre_create(**kwargs)
        kwargs.setdefault("type", "assignment")

        if kwargs.get("resource"):
            assert isinstance(kwargs["resource"], EntityCollection), \
                "Argument 'resource' expects a value of EntityCollection[User]"
            assert len(kwargs["resource"]) == 1, "'resource' can only be an EntityCollection with one entity."

            kwargs["resource"] = list(kwargs["resource"].values())[0].ftrack_entity

        return kwargs


class ObjectType(Entity):
    projections = ["id"]


class Context(Entity):
    projections = ["id"]


class AssetType(Entity):
    projections = ["id", "name"]


class Event(Entity):
    projections = ["id", "action", "user", "data"]

    relationship = Relationship()

    @Criteria.supported_targets(Entity)
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")

    @Criteria.supported_targets()
    def by_action(self, target, *actions):
        target = target or TargetRelation()
        return build_partial_query(target, actions, "action")

    @Criteria.supported_targets()
    def by_data(self, target, *datas):
        target = target or TargetRelation()
        return build_partial_query(target, datas, "data")


class TypedContextLink(Entity):
    projections = ["id", "from_id", "to_id"]


class TaskTypeSchema(Entity):
    projections = ["id", "types.name"]


class Type(Entity):
    projections = ["id", "name"]


class Metadata(Entity):
    projections = ["id", "key", "value"]


class NoteCategory(Entity):
    projections = ["id", "name"]

    @Criteria.supported_targets()
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")


class Note(Entity):
    projections = ["id", "content"]
    relationship = Relationship()

    # FIX: why do we need to set a parent if the precreate set
    #  the no_parent flag?
    relationship.parent = "parent_id"

    def pre_create(self, **kwargs):
        super(Note, self).pre_create(**kwargs)
        # assert "recipients" in kwargs, "recipients is a required keyword argument"

        author = kwargs.get("author")
        if author:
            assert isinstance(author, EntityCollection) and len(author) == 1,\
                "`author` must be a single element EntityCollection."
            author = list(author.values())[0]
            assert isinstance(author, User), "`author` collection must contain an element of type User."
            kwargs["author"] = author.ftrack_entity

        category = kwargs.get("category")
        if category:
            assert isinstance(category, EntityCollection) and len(category) == 1, \
                "`category` must be a single element EntityCollection."
            category = list(category.values())[0]
            # Querying a NodeCategory trough the Ftrack API returns a
            # dynamic NoteLabel class, hence this type is also valid
            assert isinstance(category, (NoteCategory, NoteLabel)), \
                "`category` collection must contain an element of type NoteCategory or NoteLabel."
            kwargs["category"] = category.ftrack_entity

        recipients = kwargs.get("recipients")
        if recipients:
            if isinstance(recipients, EntityCollection):
                if recipients.entity_type == User:
                    kwargs["recipient_resource_ids"] = recipients.id
                    kwargs["recipients"] = []
                elif recipients.entity_type == Recipient:
                    kwargs["recipients"] = [_.ftrack_entity for _ in recipients.values()]
                else:
                    raise NotImplementedError(
                        "We can only handle EntityCollection[User] or EntityCollection[Recipient] as recipients."
                    )
            else:
                raise AssertionError("`recipients`must be of type EntityCollection.")
        else:
            # TODO: handle replies??
            pass

        kwargs["no_parent"] = True
        return kwargs

    @Criteria.supported_targets(User, NoteCategory)
    def by_name(self, target, *names):
        target = target or TargetRelation()
        if target is self.relationship[User]:
            return build_partial_query(target, names, "username")
        return build_partial_query(target, names, "name")

    @Criteria.supported_targets(Entity)
    def by_id(self, target, *ids):
        # TODO: does not support the usual type-check for "parent" targets
        # TODO: check against type 'TypedContext' (+AssetVersion / Project? )
        if target and target.relation.startswith('parent['):
            query = "parent_id in ({})".format(", ".join(ids))
            # TODO: enable when ftrack api supports the query for "parent_type":
            # query += " and parent_type is {}".format(target.relation.split("[")[1].rstrip("]"))
            return query
        return super(Note, self).by_id(target, *ids)


class ReviewSession(Entity):
    projections = ["id", "name"]
    relationship = Relationship()

    def pre_create(self, **kwargs):
        super(ReviewSession, self).pre_create(**kwargs)

        assert "name" in kwargs, "name is a required keyword argument"
        assert "project" in kwargs, "project is a required keyword argument"
        kwargs["no_parent"] = True
        return kwargs

    def create(self, name, project_schema, **kwargs): pass

    @Criteria.supported_targets(Entity)
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")


class ReviewSessionObject(Entity):
    projections = ["id", "name"]
    relationship = Relationship()
    relationship.parent = "review_session"

    def pre_create(self, **kwargs):
        super(ReviewSessionObject, self).pre_create(**kwargs)

        if "asset_version" in kwargs:
            assert isinstance(kwargs["asset_version"], EntityCollection), "Argument 'asset_version' expects a value of EntityCollection[AssetVersion]"
            kwargs["asset_version"] = list(kwargs["asset_version"].values())[0].ftrack_entity
        return kwargs


class ReviewSessionInvitee(Entity):
    projections = ["id", "name", "email"]
    relationship = Relationship()
    # TEST: Check if this is needed or how it is actually used.
    relationship.parent = "review_session"

    def pre_create(self, **kwargs):
        super(ReviewSessionInvitee, self).pre_create(**kwargs)
        return kwargs


class ReviewSessionObjectStatus(Entity):
    projections = ["id", "status"]


class Recipient(Entity):
    # recipients are only used in notes and thus we can simply project
    # everything
    projections = ["resource_id", "note", "note_id", "recipient", "user"]

    relationship = Relationship()


class EntitySetting(Entity):
    pass


class Group(Entity):
    projections = ["id", "name"]

    @Criteria.supported_targets()
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")


class Job(Entity):
    projections = ["id", "status", "data"]

    relationship = Relationship()

    @Criteria.supported_targets(Entity)
    def by_name(self, target, *names):
        if not target:
            query = "("
            for name in names:
                query += "(data any (key is name and value like \"{}\")) or ".format(name)
            query = query[:-4] + ")"
        elif target is self.relationship[User]:
            query = build_partial_query(target, names, "username")
        else:
            query = build_partial_query(target, names, "name")
        return query

    @Criteria.supported_targets()
    def by_data(self, target, *dictionaries):
        query = "("
        for _dict in dictionaries:
            query += "("
            for key, value in _dict.items():
                if target:
                    query += "{}.".format(target.relation)
                query += "data any (key like \"{}\" and value like \"{}\") and ".format(key, value)
            query = query[:-5] + ") or "
        query = query[:-4] + ")"
        return query

    @Criteria.supported_targets()
    def by_status(self, target, *status):
        target = target or TargetRelation()
        return build_partial_query(target, status, "status")

    @Criteria.supported_targets()
    def by_lifespan(self, target, start=None, end=None):
        query = ""
        if target:
            implicit_target = "{}.".format(target.relation)
        else:
            implicit_target = ""

        if start:
            query += "{}created_at >= \"{}\"".format(implicit_target, start)
        if start and end:
            query += " and "
        if end:
            query += "{}finished_at <= \"{}\"".format(implicit_target, end)

        return query

    @Criteria.supported_targets()
    def by_creation_date(self, target, creation_date):
        query = ""
        if target:
            implicit_target = "{}.".format(target.relation)
        else:
            implicit_target = ""

        query += "{}created_at like \"{}%\"".format(implicit_target, creation_date.format("YYYY-MM-DD"))

        return query

    @Criteria.supported_targets()
    def by_finish_date(self, target, finish_date):
        query = ""
        if target:
            implicit_target = "{}.".format(target.relation)
        else:
            implicit_target = ""

        query += "{}finished_at like \"{}%\"".format(implicit_target, finish_date.format("YYYY-MM-DD"))

        return query


class Membership(Entity):
    projections = ["id", "group"]


class Location(Entity):
    projections = ["id", "name"]

    @Criteria.supported_targets()
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")


class List(Entity):
    projections = ["id", "name"]

    relationship = Relationship()

    @Criteria.supported_targets(Project, ListCategory)
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")


class TypedContextList(List):
    pass


class AssetVersionList(List):

    def pre_create(self, **kwargs):
        """ pre process before create() call

        Validates and modifies passed keyword arguments from create()

        Args:
            **kwargs (): incoming keyword arguments from create(**kwargs)

        Returns:

        """
        super(AssetVersionList, self).pre_create(**kwargs)

        assert "project" in kwargs, "project is a required keyword argument"
        assert len(kwargs["project"]) == 1, "Only one project can be linked to an AssetVersionList. You provided {}.".format(
            len(kwargs["project"])
        )
        assert isinstance(kwargs["project"], EntityCollection), "Argument 'project' expects a value of EntityCollection[Project]"

        kwargs["project"] = list(kwargs["project"].values())[0].ftrack_entity

        assert "category" in kwargs, "category is a required keyword argument"
        assert len(kwargs["category"]) == 1, "Only one category can be linked to an AssetVersionList. You provided {}.".format(
            len(kwargs["category"])
        )
        assert isinstance(kwargs["category"], EntityCollection), "Argument 'category' expects a value of EntityCollection[ListCategory]"

        kwargs["category"] = list(kwargs["category"].values())[0].ftrack_entity

        return kwargs

    # only for autocompletion
    def create(self, task, **kwargs): pass


class ListCategory(Entity):
    projections = ["id", "name"]

    relationship = Relationship()

    @Criteria.supported_targets()
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")


class WorkflowSchema(Entity):
    pass


class NoteLabel(Entity):
    projections = ["id", "name"]

    @Criteria.supported_targets(Entity)
    def by_name(self, target, *names):
        target = target or TargetRelation()
        return build_partial_query(target, names, "name")


class Resource(Entity):
    pass


# automatically declare classes for all TypedContext objects
for _type in OBJECT_TYPES.types:
    if _type not in locals() or issubclass(locals().get(_type, None), ForwardDeclaration):
        locals()[_type] = type(str(_type), (TypedContext,), {"projections": ['id', 'name']})
